/**
 * frame.c
 *
 * Author: Christian Jung <chris@tapage.de>
 * Date  : 2009-06-06
 */

#include "frame.h"

/**
 * Receive a byte from file descriptor fd and optionally compute the new CRC.
 * The byte is written to the one byte buffer pointed by dest.
 *
 * No CRC will be computed if crc is a null pointer.
 *
 * The data will be read from fd but not saved anywhere if dest is a null
 * pointer. The CRC is computed anyway.
 *
 * Returns the number of bytes read or -1 if an error occured.
 */
static ssize_t _receive_byte( const int fd, uint16_t *crc, uint8_t *dest ) {
	uint8_t t;

	if ( read( fd, &t, sizeof(t) ) != sizeof(t) ) return -1;

	if ( t == 0x10 ) {
		if ( read( fd, &t, sizeof(t) ) != sizeof(t) ) return -1;
		t &= 0x7f;
	}
	
	if ( crc != (uint16_t *)0 ) {
		*crc = crc16_byte( *crc, t );
	}

	if ( dest != (uint8_t *)0 ) {
		*dest = t;
	}
	return 1;
}

/**
 * Receive a word (16 bit) from file descriptor fd and optionally compute the
 * new CRC. The word is written to the 16 bit wide buffer pointed by dest.
 *
 * No CRC will be computed if crc is a null pointer.
 *
 * The data will be read from fd but not saved anywhere if dest is a null
 * pointer. The CRC is computed anyway.
 *
 * Returns the number of bytes read or -1 if an error occured.
 */
static ssize_t _receive_word( const int fd, uint16_t *crc, uint16_t *dest ) {
	uint8_t hi, lo;

	if ( _receive_byte( fd, crc, &hi ) != 1 ) return -1;
	if ( _receive_byte( fd, crc, &lo ) != 1 ) return -1;

	*dest = (((uint16_t)hi) << 8) | (uint16_t)lo;

	return 2;
}

ssize_t frame_receive( const int fd, uint8_t *buffer, size_t length ) {
	uint8_t t = 0;
	uint16_t payload_size = 0;
	uint16_t computed_crc = 0;
	uint16_t crc = 0;
	int i = 0;

	/* read first byte raw from fd (no DLE-decoding wanted): */
	if ( read( fd, &t, sizeof(t) ) != sizeof(t) || t != 0x02 ) return -1;

	computed_crc = crc16_init( 0xffff );

	if ( _receive_word( fd, &computed_crc, &payload_size ) != 2 ) return -1;
	
	if ( payload_size > length) return -1;

	for ( i = 0; i < payload_size; i++ ) {
		if ( _receive_byte( fd, &computed_crc, buffer++ ) != 1 ) return -1;
	}
	
	if ( _receive_word( fd, 0, &crc ) != 2 ) return -1;

	if ( computed_crc != crc ) return -1;
	return (ssize_t)payload_size;
}

static ssize_t _send_byte( const int fd, uint16_t *crc, uint8_t src ) {
	uint8_t t = src;

	if ( t == 0x02 || t == 0x10 ) {
		t = 0x10;
		if ( write( fd, &t, sizeof(t) ) != sizeof(t)) return -1;
		t = src | 0x80;
	}
	
	if ( write( fd, &t, sizeof(t) ) != sizeof(t) ) return -1;

	if ( crc != (uint16_t *)0 ) {
		*crc = crc16_byte( *crc, src );
	}

	return 1;
}

static ssize_t _send_word( const int fd, uint16_t *crc, uint16_t src ) {
	uint8_t hi = (uint8_t)(src >> 8);
	uint8_t lo = (uint8_t)(src & 0xff);

	if ( _send_byte( fd, crc, hi ) != 1 ) return -1;
	if ( _send_byte( fd, crc, lo ) != 1 ) return -1;

	return 2;
}

ssize_t frame_send( const int fd, const uint8_t *buffer, const size_t length ) {
	uint8_t t = 0;
	uint16_t crc = 0;
	int i = 0;

	if ( length > 0xffff ) return -1;

	/* send 0x02 (STX) directly (no DLE-encoding wanted) */
	t = 0x02;
	if ( write( fd, &t, sizeof(t) ) != sizeof(t) ) return -1;

	crc = crc16_init( 0xffff );

	if ( _send_word( fd, &crc, (uint16_t)(length & 0xffff) ) != 2 ) return -1;

	for ( i = 0; i < length; i++ ) {
		if ( _send_byte( fd, &crc, *buffer++ ) != 1 ) return -1;
	}
	
	if ( _send_word( fd, 0, crc ) != 2 ) return -1;
	
	return (ssize_t)length;
}

