/**
 * lcp100.c
 *
 * Demo implementation of LCP 100 protocol.
 *
 * Author: Christian Jung <chris@tapage.de>
 * Date  : 2009-05-06
 */

#include "lcp100.h"


/**
 * Open the device and initialize it (38400, 8 bit/character, 1 stop bit, even
 * paritiy). This method initializes the CRC 16 table too.
 *
 * @device  path to special file (e.g. "/dev/ttyUSB0")
 *
 * @return  file descriptor or value < 0 on error
 */
int lcp100_open( const char *device ) {
	int fd = 0;
	struct termios termios_s;
	int major, minor;

	fd = open( device, O_RDWR );
	if ( fd < 0 ) return -1;
	if ( tcgetattr( fd, &termios_s ) != 0 ) return -1;

	cfmakeraw( &termios_s );
	cfsetispeed( &termios_s, B38400 );
	cfsetospeed( &termios_s, B38400 );

	/* reset character size, set one stop bit, reset odd parity: */
	termios_s.c_cflag     &= ~( CSIZE | CSTOPB | PARODD );
	/* set character size to 8, enable parity generation: */
	termios_s.c_cflag     |=  ( CS8 | PARENB );

	/* set timeout to 10 seconds: */
	termios_s.c_cc[VMIN]   = 0;
	termios_s.c_cc[VTIME]  = 100;
	
	if ( tcsetattr(fd, TCSADRAIN, &termios_s ) != 0 ) return -1;

	/* initialize CRC 16 table */
	crc16_init_table( LCP100_CRC16_POLYNOM );

	/* check if this is a LCP100: */
	if ( lcp100_version( fd, &major, &minor ) != 0 ) return -1;

	return fd;
}


/**
 * Initialize the display.
 *
 * @fd      file descriptor of device
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_init( const int fd ) {
#ifdef LCP100_WORKAROUND_CLEAR
	uint8_t buffer[ LCP100_PAGE_SIZE ];
	int i = 0;
	memset( buffer, 0xff, sizeof(buffer) );
	for ( i = LCP100_NUMBER_PAGES - LCP100_PAGES_PER_IMAGE; 
			i < LCP100_NUMBER_PAGES; i++ ) {
		lcp100_page_upload( fd, i, buffer, sizeof(buffer) );
	}
	lcp100_test_image( fd );
#endif /* LCP100_WORKAROUND_CLEAR */
	lcp100_backlight( fd, LCP100_BACKLIGHT_OFF );
	lcp100_clear( fd );
	lcp100_backlight( fd, LCP100_BACKLIGHT_ON );

	return 0;
}


/**
 * Send the given command with its parameters and receive the response. The
 * response has to be only the ACK or NAK of the command. No additional data
 * may be supllied. If no parameters are neccessary for the command
 * 'parameters' should be a null pointer and/or 'length' should be 0.
 *
 * @fd          file descriptor of device
 * @command     LCP 100 command (LCP100_CMD_*)
 * @parameters  pointer to buffer containing additional parameters or
 *              NULL-pointer, if command does not have any parameters
 * @length      length of parameters buffer
 *
 * @return      0 upon success, < 0 in case of error
 */
static int _simple_command( const int fd, uint8_t command, const uint8_t *parameters, size_t length ) {
	int result = 0;
	uint8_t response = LCP100_NAK;
	uint8_t buffer[LCP100_MAX_PACKET_SIZE];

	if ( length > LCP100_MAX_PACKET_SIZE - sizeof(command) ) return -1;

	buffer[0] = command;
	if ( parameters != 0 && length > 0 ) {
		memcpy( buffer + 1, parameters, length );
	} else {
		length = 0;
	}

	result = frame_send( fd, buffer, sizeof(command) + length );
	if ( result < 0 ) return result;

	result = frame_receive( fd, &response, sizeof(response) );
	if ( result != 1 ) return ( result < 0 ? result : -1 );

	return ( response == LCP100_ACK ? 0 : -1 );
}


/**
 * Clears the display.
 *
 * @fd      file descriptor of device
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_clear( const int fd ) {
#ifdef LCP100_WORKAROUND_CLEAR
	int i;
	for ( i = 4032; i < 4096; i++ ) {
		if ( lcp100_page_show( fd, i, 512 ) != 0 ) return -1;
	}
	return 0;
#else
	return  _simple_command( fd, LCP100_CMD_CLEAR, 0, 0 ); 
#endif /* LCP100_WORKAROUND_CLEAR */
}


/**
 * Show generated test image.
 *
 * @fd      file descriptor of device
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_test_image( const int fd ) {
	return _simple_command( fd, LCP100_CMD_TEST, 0, 0 );
}


/**
 * Switch on/off backlight.
 *
 * @fd      file descriptor of device
 * @status  LCP_BACKLIGHT_OFF = off, LCP_BACKLIGHT_ON = on
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_backlight( const int fd, const int status ) {
	uint8_t parameter = ( 
		status ? LCP100_BACKLIGHT_ON : LCP100_BACKLIGHT_OFF );
	return _simple_command( fd, LCP100_CMD_BACKLIGHT, &parameter, sizeof(parameter) );
}


/**
 * Upload a page to the memory of the LCP 100.
 *
 * @fd      file descriptor of device
 * @page    pointer to page data
 * @length  bytes of page (have to be LCP100_PAGE_SIZE)
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_page_upload( const int fd, uint16_t index, const uint8_t *page, size_t length ) {
	uint8_t buffer[ LCP100_PAGE_SIZE + 2 ];

	if ( index >= LCP100_NUMBER_PAGES || page == 0 ||
	 		length != LCP100_PAGE_SIZE ) return -1;

	index = htons(index);
	memcpy( buffer, &index, sizeof(index) );
	memcpy( buffer + sizeof(index), page, length );
	return _simple_command( fd, LCP100_CMD_SAVE_PAGE, buffer, length + sizeof(index) );
}


/**
 * Display the page with the given index on the display.
 *
 * @fd      file descriptor of device
 * @index   index of length
 * @length  number bytes to be displayed
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_page_show( const int fd, uint16_t index, uint16_t length ) {
	uint16_t buffer[2];

	if ( index >= LCP100_NUMBER_PAGES || length > LCP100_PAGE_SIZE ) return -1;

	buffer[0] = htons( index );
	buffer[1] = htons( length );

	return _simple_command( fd, LCP100_CMD_SHOW_PAGE, (uint8_t *)buffer, sizeof(buffer) );
}


/**
 * Send command directly to the ST7637 display controller.
 *
 * @fd      file descriptor of device
 * @command command
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_st7637_cmd( const int fd, const uint8_t command ) {
	return _simple_command( fd, LCP100_CMD_ST7637_CMD, &command, sizeof(command) );
}


/**
 * Send data directly to the ST7637 display controller.
 *
 * @fd      file descriptor of device
 * @data    data
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_st7637_data( const int fd, const uint8_t data ) {
	return _simple_command( fd, LCP100_CMD_ST7637_DATA, &data, sizeof(data) );
}


/**
 * Get firmware version of LCP 100 (major- and minor number).
 *
 * @fd      file descriptor of device
 * @major   pointer to integer for the major number
 * @minor   pointer to integer for the minor number
 *
 * @return  0 upon success, < 0 in case of error
 */
int lcp100_version( const int fd, int *major, int *minor ) {
	uint8_t cmd = LCP100_CMD_VERSION;
	uint8_t response[2];
	int result;

	memset( response, 0, sizeof(response) );

	result = frame_send( fd, &cmd, sizeof(cmd) );
	if ( result < 0 ) return result;

	result = frame_receive( fd, response, sizeof(response) );
	if ( result < 0 ) return result;

	*major = (int)response[0];
	*minor = (int)response[1];

	return 0;
}
